# Events
#
# By Gino Wuytjens, 2012

import logging
from collections import deque

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

# Event Types
SESSION_START = 'SESSION_START'
SESSION_STOP = 'SESSION_STOP'
ROUND_START = 'ROUND_START'
ROUND_STOP = 'ROUND_STOP'
PLAYER_JOIN = 'PLAYER_JOIN'
PLAYER_LEAVE = 'PLAYER_LEAVE'
MAP_SWITCH = 'MAP_SWITCH'


class Event():

    def __init__(self, etype, params=None, tickid=False):
        self._type = etype  # The event type, as defined above
        self._params = params  # Parameter(s) for the event
        self._tickid = tickid  # An identification for the tick during which the event is created, see EventQueue.clean

    def getType(self):
        return self._type

    def getParams(self):
        return self._params

    def getValue(self, valuename):
        return self._params[valuename]


class EventQueue():

    def __init__(self):
        self.deque = deque()
        self._tickid = False

    def push(self, etype, params=None):
        event = Event(etype, params, self._tickid)
        self.deque.append(event)

        logger.debug("Pushing event %s with parameter %s", etype, params)

    def get(self, etype=None):
        if etype is None:
            return self.deque.popleft()
        else:
            for event in self.deque:
                if event.getType() == etype:
                    self.deque.remove(event)
                    return event

    def getAll(self):
        events = []
        for event in self.deque:
            events.append(event)
        self.deque.clear()
        self.clean()
        return events

    def clean(self):
        """Removes events that were pushed before the previous call to this function and
        were not accessed between this call and the previous.

        This allows easy enforcing of the Statecharts semantic rule that events that were not processed must be discarded.
        By tagging each event with an identifier which is changed on each call of this function, we can find the events that can safely be discarded."""

        self._tickid = ~self._tickid

        while len(self.deque) > 0 and self.deque[0]._tickid == self._tickid:
            logger.debug("Discarding unaccessed event %s", self.deque[0].getType())
            self.deque.popleft()
