# Map
#
# By Gino Wuytjens, 2012

from .. import sprites
import pygame
import os
import json
import random
import logging
import pygame.gfxdraw
from base.sprites import spritesheet
from base.sprites.pickup import FuelBarrel, RepairShop

logger = logging.getLogger(__name__)


class Map:
    """Map class with on which pawns can be placed."""

    spritesheet = None
    blueprint = None

    area = None
    surface = None

    def __init__(self, game, mapfile):
        """Creates a game map from its file.
        The parameter mapFile is the configuration of the map. For an example, see the maps included in the data/maps folder."""

        logger.info('Creating map')

        cfg = json.load(file("data/maps/" + mapfile))
        self.config = cfg

        self.game = game
        self.spritesheet = spritesheet.Spritesheet(cfg["sprites"])
        self.blueprint = pygame.image.load(os.path.join('data/maps', cfg["blueprint"])).convert()
        self.blueprint.set_colorkey(pygame.Color('0xEE00FF'))
        self.collisionMask = pygame.mask.from_surface(self.blueprint)
        self.area = self.blueprint.get_size()
        self.walls = self._determineWalls()

        if cfg["image"]["type"] == "file":  # Load this file before rendering
            self.image = pygame.image.load(os.path.join('data/maps', cfg["image"]["file"]))

        self.ambient = cfg["ambient"]
        self.surface = self.render()
        game.luminosity.setAmbient(self.ambient)
        self.loadPickups()

    def _determineWalls(self):
        maskSize = self.collisionMask.get_size()
        rects = []

        for y in range(0, maskSize[1]):  # For each row of the mask
            leftBound = None
            for x in range(0, maskSize[0]):  # Find the connected pieces
                if leftBound == None and self.collisionMask.get_at((x, y)):
                    leftBound = x
                elif leftBound != None and (self.collisionMask.get_at((x, y)) == 0 or x == maskSize[0] - 1):

                    def growableRect(rect):
                        return rect.bottom == y - 1 and rect.left == leftBound and rect.right == x

                    # Find rectangle to union with
                    rectsToGrow = filter(growableRect, rects)
                    if len(rectsToGrow) > 0:
                        for rect in rectsToGrow:
                            rect.height = rect.height + 1
                    else:
                        rects.append(pygame.Rect(leftBound, y, x - leftBound, 1))

                    leftBound = None

        xtiles = self.area[0] / self.spritesheet.tilesize
        ytiles = self.area[1] / self.spritesheet.tilesize
        logger.info("Map consists of %s wall rectangles and %s nodes", len(rects), xtiles * ytiles)

        # Debug: prints the nodes that are inaccessible due to a wall if the nodes are represented as a tiled self.spritesheet.tilesize square grid
        dnode = pygame.Rect(0, 0, self.spritesheet.tilesize, self.spritesheet.tilesize)
        for y in range(0, ytiles):
            row = ""
            for x in range(0, xtiles):
                dnode.left = x * self.spritesheet.tilesize
                dnode.top = y * self.spritesheet.tilesize
                if dnode.collidelist(rects) == -1:
                    row += " "
                else:
                    row += "#"
            logger.debug(row)
        return rects

    def unload(self):
        for pickup in self.pickups:
            pickup.vanish()
            del pickup

    def render(self):
        """Returns a Surface with the rendered map."""

        if self.config["image"]["type"] == "tiled":
            surf = self.spritesheet.makeTiledSurface(self.config["image"]["background"], self.area)
            walls = pygame.Surface(self.area)
            walls.blit(self.blueprint, (0, 0))
            walls.blit(self.spritesheet.makeTiledSurface(self.config["image"]["walls"], self.area), (0, 0), None, pygame.BLEND_MULT)
            walls.set_colorkey(pygame.Color('0x000000'))
            surf.blit(walls, (0, 0))
            surf.fill((self.ambient, self.ambient, self.ambient, self.ambient), special_flags=pygame.BLEND_MULT)
            return surf
        else:
            return self.image

    def collision(self, sprite, findTargets=True):
        """Check for collision with the map and its content.
        If findTargets is True, the function will return a list of colliding items."""

        collision = 'none'

        # Use the sprite's mask for collision detection with the map's collision mask
        spriteMask = getattr(sprite, "mask", None)
        if spriteMask == None:  # The sprite did not define a mask
            spriteMask = pygame.mask.from_surface(sprite.image)  # Create it from the sprite's image

        # Check for overlap between collision mask and sprite mask
        if self.collisionMask.overlap(spriteMask, (sprite.rect.left, sprite.rect.top)):  # Offset given by sprite's position in the map
            collision = 'wall'

        def nonCollidingSprites(thing):  # filter out myself and my owner if i'm a projectile
            #sameSprite = thing is sprite
            allies = isinstance(sprite, sprites.sprite.OwnedObject) and sprite.isMine(thing)
            nonCollideable = not isinstance(thing, sprites.sprite.CollideableSprite)

            return not (allies or nonCollideable)

        # Use spritecollide to find colliding visual representations of items on the map
        colliders = filter(nonCollidingSprites, pygame.sprite.spritecollide(sprite, self.game.playerSprites, False, pygame.sprite.collide_mask))
        if (len(colliders) > 0):
            collision = 'player'

        return (collision, colliders)

    def getWalls(self):
        """Returns the walls in the map as a list of pygame.Rect"""

        return self.walls

    def getWallMask(self):
        """Returns the walls in the map as a pygame.Mask"""

        return self.collisionMask

    def spawn(self, game, player):
        """Spawn a player somewhere on the map."""
        # TODO Uses any spawn restrictions placed on the player in the game's session.

        # Determine area around player to check, we just assume the ordinary tile size is ok
        width = self.spritesheet.tilesize
        height = width

        # Greedy sampling of the map area
        tries = 1
        x = random.randint(0 + width / 2, self.area[0] - width / 2)
        y = random.randint(0 + height / 2, self.area[1] - height / 2)
        while player.tank.setPos((x, y))[0] != "ok":  # self.collision(player.tank) != (False, [])):#(x, y), (width+height))): # call canstandhere with radius equal to 2 times the avg of width and height
            x = random.randint(0 + width / 2, self.area[0] - width / 2)
            y = random.randint(0 + height / 2, self.area[1] - height / 2)
            tries += 1
            if tries > 100:  # give up and return last, but unsafe, spawnpoint
                logger.error('Failed to create safe spawnpoint after 100 tries')
                break

        logger.info("Spawning %s at %s,%s", player, x, y)
        return (x, y)

    def playersAt(self, position, radius=1):
        srf = pygame.Surface(self.area)
        srf.set_colorkey((0, 0, 0))
        pygame.draw.circle(srf, (255, 255, 255), position, radius)
        mask = pygame.mask.from_surface(srf)

        foundPlayers = []

        for player in self.game.session.getLivePlayers():
            pmask = pygame.mask.from_surface(player.tank.image)
            if mask.overlap(pmask, (player.tank.rect.left, player.tank.rect.top)) != None:
                foundPlayers.append(player)

        return foundPlayers

    def loadPickups(self):
        """Loads all the pickups defined for this map from the map's config file."""

        self.pickups = []
        for pickup in self.config["pickups"]:
            if pickup["type"] == "fuel":
                self.pickups.append(FuelBarrel(self.spritesheet, self.game, pickup["pos"]))
            elif pickup["type"] == "health":
                self.pickups.append(RepairShop(self.spritesheet, self.game, pickup["pos"]))

    def tick(self, time_passed):
        """Advance the time by time_passed.
        This function will take care of any pickups present on the map."""

        for pickup in self.pickups:
            pickup.tick(time_passed)
