# Component classes
#
# By Gino Wuytjens, 2012

import code
import json
import logging
import pygame
import math
import random
from collections import deque

import networkx as nx

from base.events import EventQueue
from base.lib import gamemath
from base import sprites
from base.sprites.luminosity import Light

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

_COMPONENT_DIRECTORY = "data/components/"
_COMPONENT_MODULE = "data.components"


class ComponentLayer:
    """Groups components in a layer, allowing attribute access to the components in the layers.
    Based on recipe 4.18 from the Python Cookbook."""

    def __init__(self, **kwds):
        """Create a component layer consisting of the components specified by the keyword parameters."""

        self.events = EventQueue()  # Event queue for events generated by the model
        self.components = kwds.values()

        for component in self.components:
            component.setLayer(self)

        self.__dict__.update(kwds)

    def event(self, event, params=None):
        """Sends an event to all component models in this layer."""

        for component in self.components:
            component.model.event(event, params)

    def start(self):
        """Put the components in their starting configurations."""

        for component in self.components:
            component.start()

    def stop(self):
        """Stops the components."""

        for component in self.components:
            component.stop()

    def get(self):
        """Returns a list of all components in this layer."""

        return self.components


class Component:
    """Component base class, contains logic to allow loading of component configuration files in the json fileformat."""

    def __init__(self, game, player, componentconfig, localVars=None):
        """Creates a base component, loads the configuration and stores it as an attribute.
        The value of the 'model' key in the resulting configuration is loaded as the behaviour model for this component. If no model key is specified,
        no behaviour model is associated with this component.

        The game parameter is used to provide the models access to the game's notion of time and the player parameter allows the component to interact with the player.

        The optional parameter localVars allows for providing the component's model to access the local variables of the component.
        Must be a dictionary of variable names mapping to objects.

        Remember to call this from any subclass!"""

        self.game = game
        self.player = player
        self._config = json.load(file(_COMPONENT_DIRECTORY + componentconfig))

        model_module_name = self.readConfigValue("model")
        if model_module_name is not None:
            model_module = __import__(_COMPONENT_MODULE + '.' + model_module_name, globals(), locals(), [model_module_name.encode('ascii', 'ignore')], 0)
            self.model = getattr(model_module, model_module_name)(code.InteractiveInterpreter(localVars))

    def start(self):
        """Set all attributes to their starting value. If there is a model attribute, the model is sent the 'START' event.
        Should be overridden by all subclasses with their own values. If overridden, call this method or send the 'START' event in the overriding method."""

        if hasattr(self, "model"):
            self.model.initModel(self.game)
            self.model.event("START")

    def stop(self):
        """Stop any activities performed by the component. Effectively 'resets' the statecharts to their default states.
        This calls initModel on the component's model if there is any. If this method is overridden, it must call initModel itself or call this method."""

        if hasattr(self, "model"):
            del self.game.schedule[self.model.pollingStation]

    def readConfigValue(self, key):
        """Read a value from the configuration. If the key did not exist, None is returned.
        This can be done manually by accessing the _config attribute, this is not recommended
        as a clear distinction between game component attributes and their metadata keeps the
        code readable and maintainable."""

        try:
            value = self._config[key]
        except Exception:
            value = None
        finally:
            return value

    def setLayer(self, layer):
        """Sets the layer this component belongs to. After the execution of this function, the component has access to a layer attribute, specifying the
        component's layer and an events attribute allowing the component to add events to the event queue of the layer."""

        logger.debug("Component %s added to layer %s", self, layer)

        self.layer = layer
        self.events = layer.events


# Sensor Layer Components #
#                         #
# # # # # # # # # # # # # #


class Health(Component):
    """Health component"""

    def __init__(self, game, player, componentconfig="health_default.cfg"):
        """Creates a Health component. The parameter componentconfig specifies the configuration file to use and defaults to health_default.cfg"""

        Component.__init__(self, game, player, componentconfig, locals())
        self.maxHealth = self.readConfigValue("maxHealth")  # The maximum health possible, should not be used outside of this component
        self.lowHealth = self.readConfigValue("lowHealth")  # The low health value for this component

    def start(self):
        """Set the health to the max health value."""

        self.health = self.maxHealth  # The current health value
        Component.start(self)

    def isBelowPct(self, value):
        """Check if the health is below some percentage of the maximum."""

        return self.health / self.maxHealth * 100 < value

    def isBelow(self, value):
        """Check if the health is below some absolute value."""

        return self.health < value


class Fuel(Component):
    """Fuel component"""

    def __init__(self, game, player, componentconfig="fuel_default.cfg"):
        """Creates a Fuel component. The parameter componentconfig specifies the configuration file to use and defaults to fuel_default.cfg"""

        Component.__init__(self, game, player, componentconfig, locals())
        self.maxFuel = self.readConfigValue("maxFuel")  # Maximum possible level of fuel
        self.lowFuel = self.readConfigValue("lowFuel")  # Low fuel value

    def start(self):
        """Set the fuel to the max fuel value."""

        self.fuel = self.maxFuel  # Current health value
        Component.start(self)

    def getMaxLevel(self):
        """Returns the maximum amount of fuel possible."""

        return self.maxFuel

    def isBelowPct(self, value):
        """Check if the fuel is below some percentage of the maximum."""

        return self.fuel / self.maxFuel * 100 < value

    def isBelow(self, value):
        """Chec if the fuel is below some absolute value."""

        return self.fuel < value


class Radar(Component):
    """Radar component, it is capable of detecting enemies/friendlies/walls in front of the object it is attached to."""

    def __init__(self, game, player, attachTo, componentconfig="radar_default.cfg"):
        """Creates a Radar component. The parameter componentconfig specifies the configuration file to use and defaults to radar_default.cfg.
        The parameter attachTo specifies the object where the radar is mounted on. This object will determine the direction the radar is scanning in."""

        Component.__init__(self, game, player, componentconfig, locals())
        self.range = self.readConfigValue("range")
        self.fov = self.readConfigValue("fov")
        self.attachedTo = attachTo

    def start(self):
        """Start the radar."""

        self.light = Light(self.game, self.attachedTo, 30, self.range, self.fov)

        self.enemies = []
        self.friendlies = []
        self.walls = []
        self.pickups = []
        Component.start(self)

    def stop(self):
        """Stop the radar. Overridden to remove the created light source."""

        Component.stop(self)
        self.light.vanish()
        del self.light

    def scanEnemies(self):
        """Returns True if an enemy is seen by the radar.
        All enemies that were seen can be accessed through the enemies attribute, which is a list of sprites."""

        visibles = pygame.sprite.spritecollide(self.light, self.game.playerSprites, False, pygame.sprite.collide_mask)
        self.enemies = filter(self.player.tank.isEnemy, visibles)
        return len(self.enemies) > 0

    def scanFriendlies(self):
        """Returns True if a friend is seen by the radar.
        All found friends can be accessed through the friendlies attribute, which is a list of sprites."""

        visibles = pygame.sprite.spritecollide(self.light, self.game.playerSprites, False, pygame.sprite.collide_mask)
        self.friendlies = filter(self.player.tank.isFriendly, visibles)
        return len(self.friendlies) > 0

    def scanWalls(self):
        """Returns True if a wall is seen by the radar.
        All found walls can be accessed through the walls attribute, which is a list of pygame.Rect"""

        walls = self.game.session.current_map.getWalls()  # All walls of the map
        visibleWallIndices = self.light.get_bounding_rect().collidelistall(walls)  # Are checked agains the light emitted by the radar
        self.walls = [walls[i] for i in visibleWallIndices]  # We save the walls we see
        return len(self.walls) > 0  # And return wether we are seeing walls or not

    def scanPickups(self):
        """Returns True if a pickup is seen by the radar.
        All pickups can be accessed through the pickups attribute, which is list of sprites.pickup.Pickup subclass instances."""

        self.pickups = pygame.sprite.spritecollide(self.light, self.game.pickups, False, pygame.sprite.collide_mask)
        return len(self.pickups) > 0


# Analyzer Layer Components #
#                           #
# # # # # # # # # # # # # # #

def _isTank(sprite):
    """Checks if the given sprite is a tank. For use in the builtin filter function."""

    return isinstance(sprite, sprites.tank.Tank)


def _isProjectile(sprite):
    """Checks if the given sprite is a projectile. For use in the builtin filter function."""

    return isinstance(sprite, sprites.projectile.Projectile)


class InSightsDetector(Component):
    """InSightsDetector component capable of checking if there are enemies directly in front of a radar.
    Usually used with a radar attached to the turret."""

    def __init__(self, game, player, radar, componentconfig="insightsdetector_default.cfg"):
        """Creates an InSightsDetector component. The parameter componentconfig specifies the configuration file and defaults to insightsdetector_default.cfg.
        The parameter radar should be exactly one instance of the Radar component class."""

        Component.__init__(self, game, player, componentconfig, locals())
        self.radar = radar
        self.tolerance = self.readConfigValue("tolerance")

    def somethingInFront(self, sprites):
        """Checks if something is directly in front of the radar. The sprites parameter is a list of sprites
        to check and tolerance specifies how far the angle may deviate from the radar attachment point's angle."""

        shortestDistance = 999
        closestSpriteInFront = None
        for sprite in sprites:
            angle = gamemath.angleSprite(self.radar.attachedTo, sprite)
            if math.fabs(angle - self.radar.attachedTo.getAngle()) < self.tolerance:
                distance = gamemath.distanceSprite(self.radar.attachedTo, sprite)
                if distance < shortestDistance:
                    shortestDistance = distance
                    closestSpriteInFront = sprite

        return closestSpriteInFront

    def tankInFront(self, sprites):
        """Checks if a tank is in front of the radar. See somethingInFront."""

        return self.somethingInFront(filter(_isTank, sprites))

    def projectileInFront(self, sprites):
        """Checks if a projectile is in front of the radar. See somethingInFront."""

        return self.somethingInFront(filter(_isProjectile, sprites))

    def enemyInFront(self):
        """Convenience method. Checks if an enemy is in front of the radar. Checks for anything that's hostile to us."""

        return self.somethingInFront(self.radar.enemies)

    def enemyTankInFront(self):
        """Convenience method. Checks if an enemy tank is in front."""

        return self.tankInFront(self.radar.enemies)

    def enemyProjectileInFront(self):
        """Convenience method. Checks if an enemy projectile is in front."""

        return self.projectileInFront(self.radar.enemies)

    def friendInFront(self):
        """Convenience method. Checks if a friend is in front of the radar. Checks for anything that's friendly to us."""

        return self.somethingInFront(self.radar.friendlies)

    def friendlyTankInFront(self):
        """Convenience method. Checks if a friendly tank is in front."""

        return self.tankInFront(self.radar.friendlies)

    def friendlyProjectileInFront(self):
        """Convenience method. Checks if a friendly projectile is in front."""

        return self.projectileInFront(self.radar.friendlies)

    def pickupInFront(self):
        """Convenience method. Checks if there is a pickup in front."""

        return self.somethingInFront(self.radar.pickups)

# Memorizer Layer Components  #
#                             #
# # # # # # # # # # # # # # # #


class PlayerTracker(Component):
    """Tracks players (enemies and friends)."""

    def __init__(self, game, player, radars, componentconfig="playertracker_default.cfg"):
        """Creates an PlayerTracker component. The parameter componentconfig specifies the configuration file.
        The parameter radars should be a list of all Radar instances the player has (unless you want to exclude some radars from the tracking mechanism)."""

        Component.__init__(self, game, player, componentconfig, locals())
        self.radars = radars

    def start(self):
        # Dictionaries with sprites as keys and a list of the form [Tracked?, Position Tuple]
        self.enemyPositions = {}
        self.friendPositions = {}
        Component.start(self)

    def nrTrackedEnemies(self):
        """Returns the amount of enemies currently tracked."""

        return sum(enemyPos[0] for enemyPos in self.enemyPositions.itervalues())

    def nrTrackedFriends(self):
        """Returns the amount of friends currently tracked."""

        return sum(friendPos[0] for friendPos in self.friendPositions.itervalues())

    def getTrackedFriends(self):
        """Returns all tracked friends."""

        return [friend for friend in self.friendPositions.keys() if self.friendPositions[friend][0]]

    def getTrackedEnemies(self):
        """Returns all tracked enemies."""

        return [enemy for enemy in self.enemyPositions.keys() if self.enemyPositions[enemy][0]]

    def getClosestTrackedFriend(self):
        return self.player.tank.getClosest(self.getTrackedFriends())

    def getClosestTrackedEnemy(self):
        return self.player.tank.getClosest(self.getTrackedEnemies())

    def refreshEnemyPositions(self):
        """Refreshes the positions of enemies."""

        for enemyPosition in self.enemyPositions.itervalues():
            enemyPosition[0] = False

        for radar in self.radars:
            for enemy in radar.enemies:
                if enemy not in self.enemyPositions.iterkeys() or self.enemyPositions[enemy][1] != enemy.getPos():
                    # Send the event PLAYERTRACKER_ENEMY_MOVED
                    self.events.push("PLAYERTRACKER_ENEMY_MOVED")
                # Update the position in our internal representation
                self.enemyPositions[enemy] = [True, enemy.getPos()]

        return self.nrTrackedEnemies()

    def refreshFriendPositions(self):
        """Refreshes the positions of friends."""

        for friendPosition in self.friendPositions.itervalues():
            friendPosition[0] = False

        for radar in self.radars:
            for friend in radar.friendlies:
                if friend not in self.friendPositions.iterkeys() or self.friendPositions[friend][1] != friend.getPos():
                    # Send the event PLAYERTRACKER_FRIEND_MOVED
                    self.events.push("PLAYERTRACKER_FRIEND_MOVED")
                # Update the position in our internal representation
                self.friendPositions[friend] = [True, friend.getPos()]

        return self.nrTrackedFriends()


class MapDiscoverer(Component):
    """Discovers the layout of the map."""

    def __init__(self, game, player, radars, componentconfig="mapdiscoverer_default.cfg"):
        """Creates an MapDiscoverer component. The parameter componentconfig specifies the configuration file.
        The parameter radars should be a list of all Radar instances the player has (unless you want to exclude some radars from the discovery mechanism)."""

        Component.__init__(self, game, player, componentconfig, locals())
        self.radars = radars

    def start(self):
        self.pickupPositions = {}
        self._createMapGraph()
        Component.start(self)

    def _createMapGraph(self):
        """Creates the graph representation of the map. This representation is a grid of pygame.Rect equal in size to our tank.
        This representation allows efficient pathfinding using the networkx package. Initially, all tiles are connected to adjacent tiles."""

        mapArea = self.game.session.current_map.area
        self.tileSize = 32

        # Uses a networkx Graph. This package provides the grid_2d_graph function to create a 2d grid Graph
        self.map = nx.grid_2d_graph(mapArea[0] / self.tileSize, mapArea[1] / self.tileSize)

        # This means that we should add diagonal edges to allow diagonally oriented waypoints.
        # We set the weight attribute of each edge to the euclidean distance between the nodes.
        # tl to br diagonal edges
        for i in range(0, (mapArea[0] / self.tileSize) - 1):  # Not the last column of nodes
            for j in range(0, (mapArea[1] / self.tileSize) - 1):  # Not the last row of nodes
                self.map.add_edge((i, j), (i + 1, j + 1), weight=math.sqrt(2))
        # tr to bl diagonal edges
        for i in range(1, mapArea[0] / self.tileSize):  # Not the first column of nodes
            for j in range(0, (mapArea[1] / self.tileSize) - 1):  # Not the last row of nodes
                self.map.add_edge((i, j), (i - 1, j + 1), weight=math.sqrt(2))

    def getNodePos(self, mapNode):
        """Returns the center of the provided mapNode. For use when using mapnodes to navigate the map."""

        return (mapNode[0] * self.tileSize + self.tileSize / 2, mapNode[1] * self.tileSize + self.tileSize / 2)

    def getPosNode(self, pos):
        """Returns the map node for the specified map coordinate."""

        return (math.floor(pos[0] / self.tileSize), math.floor(pos[1] / self.tileSize))

    def getRandomNode(self):
        """Returns a random node of the map graph."""

        return random.choice(self.map.nodes())

    def updatePickupPositions(self):
        """Updates the positions of pickups as they are seen by the radars."""

        for radar in self.radars:
            for pickup in radar.pickups:
                self.pickupPositions[pickup] = pickup.getPos()

    def updateWalls(self):
        """Updates the position of walls as they are seen by the radars.
        Returns the number of map nodes deleted as a consequence of seeing new map walls."""

        allVisibleWalls = []
        for radar in self.radars:
            allVisibleWalls.extend(list(radar.walls))

        # Using a pygame.Rect - for the collision detection - we go over all the map nodes
        # and test if they overlap with any walls we can see. If they do, they are removed.
        nodeRect = pygame.Rect(0, 0, self.tileSize, self.tileSize)
        removedNodes = 0
        nodesToRemove = []
        for mapNode in self.map.nodes():
            nodeRect.left = mapNode[0] * self.tileSize
            nodeRect.top = mapNode[1] * self.tileSize
            collider = nodeRect.collidelist(allVisibleWalls)
            if collider != -1:
                nodesToRemove.append(mapNode)
                removedNodes += 1

        if len(nodesToRemove) > 0:
            logger.debug("Removing map node(s) %s", mapNode)

            for node in nodesToRemove:
                self.map.remove_node(node)  # Remove the node itself
                touchingDiagonalEdges = [((node[0], node[1] - 1), (node[0] + 1, node[1])), \
                                         ((node[0], node[1] - 1), (node[0] - 1, node[1])), \
                                         ((node[0], node[1] + 1), (node[0] + 1, node[1])), \
                                         ((node[0], node[1] + 1), (node[0] - 1, node[1]))]
                self.map.remove_edges_from(touchingDiagonalEdges)  # Remove all diagonals between neighbor nodes that overlap with the corners of the map tile the node represents

            # Debug output
            allNodes = self.map.nodes()
            for y in range(0, 480 / self.tileSize):
                row = ""
                for x in range(0, 640 / self.tileSize):
                    if (x, y) in allNodes:
                        row += " "
                    else:
                        row += "#"
                logger.debug("Learned map: %s", row)

        return removedNodes

    def knowFuelPosition(self):
        """Returns True if there is at least one fuel pickup known."""

        for pickup in self.pickupPositions.iterkeys():
            if pickup.__class__.__name__ == "FuelBarrel":
                return True

        return False

    def knowRepairPosition(self):
        """Returns True if there is at least one repair pickup known."""

        for pickup in self.pickupPositions.iterkeys():
            if pickup.__class__.__name__ == "RepairShop":
                return True

        return False

    def repairShops(self):
        """Returns a list of all RepairShop pickups found."""

        return [pickup for pickup in self.pickupPositions.iterkeys() if pickup.__class__.__name__ == "RepairShop"]

    def fuelBarrels(self):
        """Returns a list of all FuelBarrel pickups found."""

        return [pickup for pickup in self.pickupPositions.iterkeys() if pickup.__class__.__name__ == "FuelBarrel"]

# Strategic Decider Layer Components  #
#                                     #
# # # # # # # # # # # # # # # # # # # #


class Pilot(Component):
    """The pilot will make the main strategic decisions.
    It is a behavior-only component, as it does not query or alter the game-state itself, but relies on events from previous layers."""

    def __init__(self, game, player, componentconfig="pilot_default.cfg"):
        """Create a new Pilot component. The parameter componentconfig specifies the configuration file."""

        Component.__init__(self, game, player, componentconfig, locals())

# Tactical Decider Layer Components #
#                                   #
# # # # # # # # # # # # # # # # # # #


class Explorer(Component):
    """The Explorer component will make decisons on how to best explore the world."""

    def __init__(self, game, player, mapdiscoverer, componentconfig="explorer_default.cfg"):
        """Create a new Explorer component. The parameter componentconfig specifies the configuration file.
        The mapdiscoverer parameter is a component of type MapDiscoverer, which allows the Explorer component to find locations to move towards."""

        Component.__init__(self, game, player, componentconfig, locals())
        self.mapdiscoverer = mapdiscoverer

    def findNewDestination(self):
        """Finds a new destination to explore."""

        # Choose a node from the existing map nodes. Do not exclude already visited nodes, because enemies can move about the map too!
        self.destinationNode = random.choice(self.mapdiscoverer.map.nodes())
        self.destination = self.mapdiscoverer.getNodePos(self.destinationNode)
        return self.destination

    def destinationExists(self):
        """Checks if the destination exists at this time."""

        return self.destinationNode in self.mapdiscoverer.map.nodes()

    def reachedDestination(self):
        """Returns true if the currently scheduled destination is reached."""

        # We are happy enough to arrive in the map node of the destination.
        myNode = self.mapdiscoverer.getPosNode(self.player.tank.getPos())
        destNode = self.mapdiscoverer.getPosNode(self.destination)
        if myNode == destNode:
            logger.debug("Arrived at destination node containing %s", self.destination)
        return myNode == destNode


class StayAlive(Component):
    """The StayAlive component will attempt to keep the player alive by healing or avoiding enemies."""

    def __init__(self, game, player, mapdiscoverer, componentconfig="stayalive_default.cfg"):
        """Create a new StayAlive component. The parameter componentconfig specifies the configuration file.
        The mapdiscoverer parameter is a component of type MapDiscoverer, which allows the StayAlive component to find healing pickups to move towards."""

        Component.__init__(self, game, player, componentconfig, locals())
        self.mapdiscoverer = mapdiscoverer
        self.shop = None

    def closestRepairShop(self):
        """Returns the closest RepairShop pickup on the map."""

        allShops = self.mapdiscoverer.repairShops()
        nrShops = len(allShops)
        if nrShops == 0:
            return None
        elif nrShops == 1:
            return allShops[0]
        else:
            shortestDistance = 99999
            closestShop = None
            for shop in allShops:
                distance = gamemath.distanceSprite(self.player.tank, shop)
                if distance < shortestDistance:
                    shortestDistance = distance
                    closestShop = shop
            return closestShop

    def reachedDestination(self):
        """Returns true if the currently scheduled destination is reached."""

        # We are happy enough to arrive in the map node of the destination.
        myNode = self.mapdiscoverer.getPosNode(self.player.tank.getPos())
        destNode = self.mapdiscoverer.getPosNode(self.destination)
        if myNode == destNode:
            logger.debug("Arrived at destination node containing %s", self.destination)
        return myNode == destNode


class StayFueled(Component):
    """The StayFueled component will attempt to keep the player fueled by finding fuel."""

    def __init__(self, game, player, mapdiscoverer, componentconfig="stayfueled_default.cfg"):
        """Create a new StayFueled component. The parameter componentconfig specifies the configuration file.
        The mapdiscoverer parameter is a component of type MapDiscoverer, which allows the StayFueled component to find fuel pickups to move towards."""

        Component.__init__(self, game, player, componentconfig, locals())
        self.mapdiscoverer = mapdiscoverer

    def closestFuelBarrel(self):
        """Returns the closest FuelBarrel pickup on the map."""

        allBarrels = self.mapdiscoverer.fuelBarrels()
        nrBarrels = len(allBarrels)
        if nrBarrels == 0:
            return None
        elif nrBarrels == 1:
            return allBarrels[0]
        else:
            shortestDistance = 99999
            closestBarrel = None
            for barrel in allBarrels:
                distance = gamemath.distanceSprite(self.player.tank, barrel)
                if distance < shortestDistance:
                    shortestDistance = distance
                    closestBarrel = barrel
            return closestBarrel

    def reachedDestination(self):
        """Returns true if the currently scheduled destination is reached."""

        # We are happy enough to arrive in the map node of the destination.
        myNode = self.mapdiscoverer.getPosNode(self.player.tank.getPos())
        destNode = self.mapdiscoverer.getPosNode(self.destination)
        if myNode == destNode:
            logger.debug("Arrived at destination node containing %s", self.destination)
        return myNode == destNode


class Engager(Component):
    """The Engager component will attempt to keep the player fueled by finding fuel."""

    def __init__(self, game, player, playertracker, mapdiscoverer, componentconfig="engager_default.cfg"):
        """Create a new Engager component. The parameter componentconfig specifies the configuration file.
        The playertracker parameter is a component of type PlayerTracker, which allows the Engager component to find enemies to engage.
        The mapdiscoverer parameter is a component of type MapDiscoverer, which allows the Engager component to find good locations to engage from."""

        Component.__init__(self, game, player, componentconfig, locals())
        self.playertracker = playertracker
        self.mapdiscoverer = mapdiscoverer

        self.keepAtDistance = self.readConfigValue("keepAtDistance")

    def chooseTarget(self):
        self.currentTarget = self.playertracker.getClosestTrackedEnemy()

    def distanceToTarget(self):
        return gamemath.distanceSprite(self.player.tank, self.currentTarget)

    def reachedDestination(self):
        """Returns true if the currently scheduled destination is reached."""

        # We are happy enough to arrive in the map node of the destination.
        myNode = self.mapdiscoverer.getPosNode(self.player.tank.getPos())
        destNode = self.mapdiscoverer.getPosNode(self.destination)
        if myNode == destNode:
            logger.debug("Arrived at destination node containing %s", self.destination)
        return myNode == destNode

# Executor Layer Components #
#                           #
# # # # # # # # # # # # # # #


class AimExecutor(Component):
    """Generates lower level events pertaining to aiming. It can aim at any position on the map."""

    def __init__(self, game, player, componentconfig="aimexecutor_default.cfg"):
        """Create a new AimExecutor component. The parameter componentconfig specifies the configuration file."""

        Component.__init__(self, game, player, componentconfig, locals())
        self.direction = None
        self.currentTarget = None

    def findDirection(self):
        """Determines the direction for the current aim target position."""

        self.direction = gamemath.angle(self.player.tank.getPos(), self.currentTarget)


class MovementExecutor(Component):
    """Generates lower level events pertaining to movement. It can move towards any (unblocked) position on the map by navigating the map nodes found by the MapDiscoverer."""

    def __init__(self, game, player, mapdiscoverer, componentconfig="movementexecutor_default.cfg"):
        """Create a new MovementExecutor component. The parameter componentconfig specifies the configuration file.
        The mapdiscoverer parameter is a component of type MapDiscoverer, allowing the executor to find a path to the desired locations."""

        Component.__init__(self, game, player, componentconfig, locals())
        self.mapdiscoverer = mapdiscoverer
        self.currentPath = None
        self.currentTarget = None

    def findPath(self, pos):
        self.currentTarget = pos
        currentNode = self.mapdiscoverer.getPosNode(self.player.tank.getPos())
        targetNode = self.mapdiscoverer.getPosNode(pos)
        try:
            path = nx.astar_path(self.mapdiscoverer.map, currentNode, targetNode, gamemath.distance)
            logger.debug("Found a path to destination: %s", path)
            self.currentPath = deque(path)
            self.currentPath.popleft()  # Remove the source node, because we are already there!
        except nx.NetworkXNoPath:
            logger.warning("No path from %s to %s", currentNode, targetNode)
            self.currentPath = None

    def haveWaypoint(self):

        return self.currentPath != None and len(self.currentPath) > 0

    def getNextWaypoint(self):
        if self.currentPath == None or len(self.currentPath) == 0:
            # We have exausted all waypoints
            return None
        else:
            # We have at least one more waypoint to move towards
            self.currentWaypoint = self.mapdiscoverer.getNodePos(self.currentPath.popleft())
            return self.currentWaypoint

    def atWaypoint(self):
        return self.mapdiscoverer.getPosNode(self.player.tank.getPos()) == self.mapdiscoverer.getPosNode(self.currentWaypoint)

# Coordinator Layer Components  #
#                               #
# # # # # # # # # # # # # # # # #


class AimingCoordinator(Component):
    """The AimingCoordinator generates the events that are understood by the Actuators.
    It will translate events from the executors into the most efficient actuator events to achieve the desired goals."""

    def __init__(self, game, player, componentconfig="aimingcoordinator_default.cfg"):
        """Create a new AimingCoordinator component. The parameter componentconfig specifies the configuration file."""

        Component.__init__(self, game, player, componentconfig, locals())

        self.currentTarget = None

    def shouldAimLeft(self):
        """Returns True if aiming at angle by turning left is most efficient."""

        return self.currentTarget - self.player.turret.getAngle() > 0

    def shouldAimRight(self):
        """Returns True if aiming at angle by turning right is most efficient."""

        return self.currentTarget - self.player.turret.getAngle() < 0

    def isStraightAhead(self):
        """Returns True if the turret is aimed at the self.currentTarget."""

        return math.fabs(self.currentTarget - self.player.turret.getAngle()) < 0.1


class SteeringCoordinator(Component):
    """The SteeringCoordinator generates the events that are understood by the Actuators.
    It will translate events from the executors into the most efficient actuator events to achieve the desired goals."""

    def __init__(self, game, player, componentconfig="steeringcoordinator_default.cfg"):
        """Create a new SteeringCoordinator component. The parameter componentconfig specifies the configuration file."""

        Component.__init__(self, game, player, componentconfig, locals())

    def findDirection(self, pos):
        self.direction = gamemath.angle(self.player.tank.getPos(), pos)

    def shouldTurnLeft(self):
        """Returns True if turning at self.direction by turning left is most efficient."""

        return self.direction - self.player.tank.getAngle() > 0

    def shouldTurnRight(self):
        """Returns True if Turning at self.direction by turning right is most efficient."""

        return self.direction - self.player.tank.getAngle() < 0

    def isStraightAhead(self):
        """Returns True if the tank is turned at the self.direction."""

        return math.fabs(self.direction - self.player.tank.getAngle()) < 0.1

    def isLeftOfTank(self, pos):
        (x, y) = pos
        (ax, ay) = self.player.tank.getAngleProjection()
        (ax, ay) = (ax * 10, -ay * 10)
        (tx, ty) = self.player.tank.getPos()
        answer = ((tx + ax) - tx) * (y - ty) - ((ty + ay) - ty) * (x - tx) < 0
        #print pos, "is left of tank at", self.player.tank.getPos(), (tx+ax,ty+ay), "?", answer, "(", ((tx + ax) - tx) * (y - ty) - ((ty + ay) - ty) * (x - tx), ")"
        return answer

    def isRightOfTank(self, pos):
        (x, y) = pos
        (ax, ay) = self.player.tank.getAngleProjection()
        (ax, ay) = (ax * 10, -ay * 10)
        (tx, ty) = self.player.tank.getPos()
        answer = ((tx + ax) - tx) * (y - ty) - ((ty + ay) - ty) * (x - tx) > 0
        #print pos, "is right of tank at", self.player.tank.getPos(), (tx+ax,ty+ay), "?", answer, "(", ((tx + ax) - tx) * (y - ty) - ((ty + ay) - ty) * (x - tx), ")"
        return answer

    #def isStraightAhead(self, pos):
    #    answer = math.fabs(gamemath.angle(self.player.tank.getPos(), pos) - self.player.tank.getAngle()) < 0.1
    #    #print "angle straight?", answer, gamemath.angle(self.player.tank.getPos(), pos) - self.player.tank.getAngle()
    #    return answer

    def isInFrontOfTank(self, pos):
        (x, y) = pos
        (ax, ay) = self.player.tank.getAngleProjection()
        (ax, ay) = (ax * 10, -ay * 10)
        (nx, ny) = (-ay, ax)
        (tx, ty) = self.player.tank.getPos()
        answer = ((tx + nx) - tx) * (y - ty) - ((ty + ny) - ty) * (x - tx) < 0
        #print pos, "is in front of tank at", self.player.tank.getPos(), (tx+nx, ty+ny), "?", answer, "(", ((tx + nx) - tx) * (y - ty) - ((ty + ny) - ty) * (x - tx), ")"
        return answer

    def isBehindTank(self, pos):
        (x, y) = pos
        (ax, ay) = self.player.tank.getAngleProjection()
        (ax, ay) = (ax * 10, -ay * 10)
        (nx, ny) = (-ay, ax)
        (tx, ty) = self.player.tank.getPos()
        answer = ((tx + nx) - tx) * (y - ty) - ((ty + ny) - ty) * (x - tx) > 0
        #print pos, "is behind tank at", self.player.tank.getPos(), (tx+nx, ty+ny), "?", answer, "(", ((tx + nx) - tx) * (y - ty) - ((ty + ny) - ty) * (x - tx), ")"
        return answer

# Actuator Layer Components #
#                           #
# # # # # # # # # # # # # # #


class Turret(Component):
    """The turret actuator will make the player's turret move."""

    def __init__(self, game, player, componentconfig="turret_default.cfg"):
        """Create a new Turret component. The parameter componentconfig specifies the configuration file."""

        Component.__init__(self, game, player, componentconfig, locals())

        self.turnSpeed = self.readConfigValue("turnSpeed")

    def start(self):
        self.turret = self.player.turret
        Component.start(self)

    def left(self, angle):
        #self.turret.setAngle(self.turret.getAngle() + self.turnSpeed)
        self.turret.turnTo(angle, self.turnSpeed)

    def right(self, angle):
        #self.turret.setAngle(self.turret.getAngle() - self.turnSpeed)
        self.turret.turnTo(angle, -self.turnSpeed)

    def straight(self):
        self.turret.setTurnSpeed(0.0)

    def shoot(self):
        self.turret.fire()


class Motor(Component):
    """The motor actuator will make the player move across the map."""

    def __init__(self, game, player, componentconfig="motor_default.cfg"):
        """Create a new Motor component. The parameter componentconfig specifies the configuration file."""

        Component.__init__(self, game, player, componentconfig, locals())

        self.speed = self.readConfigValue("speed")
        self.turnSpeed = self.readConfigValue("turnSpeed")

    def start(self):
        self.tank = self.player.tank
        Component.start(self)

    def forward(self):
        """Convenience method, moves the player forwards."""

        self.tank.setSpeed(self.speed)

    def backward(self):
        """Convenience method, moves the player backwards."""

        self.tank.setSpeed(-self.speed)

    def idle(self):
        """Convenience method, stops player movement."""

        self.tank.setSpeed(0)

    def left(self, angle):
        #self.tank.setAngle(self.tank.getAngle() + self.turnSpeed)
        self.tank.turnTo(angle, self.turnSpeed)

    def right(self, angle):
        #self.tank.setAngle(self.tank.getAngle() - self.turnSpeed)
        self.tank.turnTo(angle, -self.turnSpeed)

    def straight(self):
        """Convenience method. Stops turning the tank."""

        self.tank.setTurnSpeed(0)
