# Controller
#
# Controllers for players
#
# Gino Wuytjens

import math
import logging

import pygame
from pygame.locals import *

from component import *

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class Controller:
    """The base Controller class, allows for any kind of implementation of the behaviour of a player."""

    def __init__(self, player):
        """Create a Controller for the provided player object."""

        self.player = player  # a pawn in the world we get to control

    def start(self):
        """Stats the controller."""

        self.active = True

    def isActive(self):
        """Returns True if the controller is active."""

        return self.active

    def stop(self):
        """Stops the controller."""

        self.active = False

    def tick(self, time_passed):
        """Called every tick, this should perform the 'living' behaviours of the associated player such as moving and shooting."""

        pass


class InputController(Controller):
    """A subclass of the base Controller class that implements player behaviour controlled by keyboard and mouse input."""

    def __init__(self, player):
        Controller.__init__(self, player)
        self.x = 0
        self.y = 0

    def tick(self, time_passed):
        if not self.active:
            return

        keypress = pygame.event.get(KEYDOWN)
        keydepress = pygame.event.get(KEYUP)
        mousepress = pygame.event.get(MOUSEBUTTONDOWN)

        fire = False

        for i in keypress:
            if i.key == K_UP:
                self.y = 1
            elif i.key == K_DOWN:
                self.y = -1
            if i.key == K_LEFT:
                self.x = -1
            elif i.key == K_RIGHT:
                self.x = 1
            elif i.key == K_SPACE:
                fire = True

        for i in keydepress:
            if i.key == K_UP or i.key == K_DOWN:
                self.y = 0
            if i.key == K_LEFT or i.key == K_RIGHT:
                self.x = 0

        for i in mousepress:
            if i.button == 1:
                fire = True

        #turretangle = math.acos((self.pawn.y-i.pos[1])/lib.gamemath.distance((self.pawn.x, self.pawn.y), (i.pos[0], i.pos[1])))#-math.pi/2
        #if self.pawn.tank.y == i.pos[1]:
        #   self.pawn.tank.y -= 0.00000001
        (mposx, mposy) = pygame.mouse.get_pos()
        if mposx == 0 or mposx == self.player.tank.x:
            mposx += 0.00000001
        if mposy == 0 or mposy == self.player.tank.y:
            mposy += 0.00000001

        turretangle = math.atan((mposx - self.player.tank.x) / (mposy - self.player.tank.y))
        if mposy < self.player.tank.y:
            turretangle = math.pi / 2.0 + turretangle
        else:
            turretangle = math.pi * 1.5 + turretangle
        self.player.tank.setTurretAngle(turretangle)

        previousAngle = self.player.tank.getAngle()
        angle = 0.0
        if self.x == 0 and self.y == 0:
            self.player.tank.setSpeed(0)
            angle = previousAngle
        else:
            self.player.tank.setSpeed(50)

        if self.x == 0 and self.y == 1:
            angle = math.pi / 2.0
        elif self.x == 0 and self.y == -1:
            angle = -math.pi / 2.0
        elif self.x == 1 and self.y == 0:
            angle = 0.0
        elif self.x == 1 and self.y == 1:
            angle = math.pi / 4.0
        elif self.x == 1 and self.y == -1:
            angle = -math.pi / 4.0
        elif self.x == -1 and self.y == 0:
            angle = math.pi
        elif self.x == -1 and self.y == 1:
            angle = 3.0 / 4.0 * math.pi
        elif self.x == -1 and self.y == -1:
            angle = -3.0 / 4.0 * math.pi

        # print self.x, self.y, angle

        if (angle != previousAngle):
            self.player.tank.setAngle(angle)

        if fire:
            self.player.turret.fire()


class ComponentController(Controller):
    """ModelController class implements Controller functionality through various components."""

    def __init__(self, player, game):
        """Creates a component controller for the provided player using the game instance provided as a means to poll the game time"""
        Controller.__init__(self, player)

        # Initialize component layers
        self.sensors = ComponentLayer(health=Health(game, player), fuel=Fuel(game, player), tankradar=Radar(game, player, player.tank, "tankradar_default.cfg"), turretradar=Radar(game, player, player.turret, "turretradar_default.cfg"))
        allRadars = [self.sensors.tankradar, self.sensors.turretradar]
        self.analyzers = ComponentLayer(insightsdetector=InSightsDetector(game, player, self.sensors.turretradar))
        self.memorizers = ComponentLayer(playertracker=PlayerTracker(game, player, allRadars), mapdiscoverer=MapDiscoverer(game, player, allRadars))
        self.stratdeciders = ComponentLayer(pilot=Pilot(game, player))
        self.tacdeciders = ComponentLayer(explorer=Explorer(game, player, self.memorizers.mapdiscoverer), stayfueled=StayFueled(game, player, self.memorizers.mapdiscoverer), stayalive=StayAlive(game, player, self.memorizers.mapdiscoverer), engager=Engager(game, player, self.memorizers.playertracker, self.memorizers.mapdiscoverer))
        self.executors = ComponentLayer(movexecutor=MovementExecutor(game, player, self.memorizers.mapdiscoverer), aimexecutor=AimExecutor(game, player))
        self.coordinators = ComponentLayer(steeringcoordinator=SteeringCoordinator(game, player), aimingcoordinator=AimingCoordinator(game, player))
        self.actuators = ComponentLayer(motor=Motor(game, player), turret=Turret(game, player))

    def start(self):
        """Starts the controller and the components it is made up of."""

        # Supertype performs bookkeeping
        Controller.start(self)

        # We start the component layers
        self.sensors.start()
        self.analyzers.start()
        self.memorizers.start()
        self.stratdeciders.start()
        self.tacdeciders.start()
        self.executors.start()
        self.coordinators.start()
        self.actuators.start()

    def stop(self):
        """Stops the controller and the components it is made up of."""

        # Supertype performs bookkeeping
        Controller.stop(self)

        # We make sure all components are in their default configuration
        self.sensors.stop()
        self.analyzers.stop()
        self.memorizers.stop()
        self.stratdeciders.stop()
        self.tacdeciders.stop()
        self.executors.stop()
        self.coordinators.stop()
        self.actuators.stop()

    def tick(self, time_passed):
        """Advance the time by time_passed and perform any required logic.
        In the case of the ComponentController this will , if active, cascade events along the layers."""

        if not self.active:
            return

        # Cascade events from the first to the last layer
        sensorEvents = self.sensors.events.getAll()
        for e in sensorEvents:
            self.analyzers.event(e.getType(), e.getParams())
            self.memorizers.event(e.getType(), e.getParams())
            self.stratdeciders.event(e.getType(), e.getParams())
            self.tacdeciders.event(e.getType(), e.getParams())
            self.executors.event(e.getType(), e.getParams())

        analyzerEvents = self.analyzers.events.getAll()
        for e in analyzerEvents:
            self.memorizers.event(e.getType(), e.getParams())
            self.stratdeciders.event(e.getType(), e.getParams())
            self.tacdeciders.event(e.getType(), e.getParams())
            self.executors.event(e.getType(), e.getParams())

        memorizerEvents = self.memorizers.events.getAll()
        for e in memorizerEvents:
            self.stratdeciders.event(e.getType(), e.getParams())
            self.tacdeciders.event(e.getType(), e.getParams())
            self.executors.event(e.getType(), e.getParams())

        stratdeciderEvents = self.stratdeciders.events.getAll()
        for e in stratdeciderEvents:
            self.tacdeciders.event(e.getType(), e.getParams())
            self.executors.event(e.getType(), e.getParams())

        tacdeciderEvents = self.tacdeciders.events.getAll()
        for e in tacdeciderEvents:
            self.executors.event(e.getType(), e.getParams())
            self.coordinators.event(e.getType(), e.getParams())
            self.actuators.event(e.getType(), e.getParams())

        executorEvents = self.executors.events.getAll()
        for e in executorEvents:
            self.coordinators.event(e.getType(), e.getParams())
            self.actuators.event(e.getType(), e.getParams())

        coordEvents = self.coordinators.events.getAll()
        for e in coordEvents:
            self.actuators.event(e.getType(), e.getParams())
