# Player
#
# By Gino Wuytjens, 2012

import pygame
import logging

from ..sprites import tank
from ..sprites import hudelements
from controller import InputController
from controller import ComponentController
from component import *

logger = logging.getLogger(__name__)


class Player:
    """A Player base class, this should be subclassed and the methods setHealth(), getHealth() need to be implemented."""

    def __init__(self, game, name="dummy", team=0):
        """Create a player that will take part in the provided game."""

        logger.info("Creating player")

        # Session and metadata
        self.session = game.session
        self.game = game
        self.createdat = pygame.time.get_ticks()
        self.deathdat = 0
        self.totaldamage = 0

        # Identity
        self.name = name
        self.team = team

        # Player is now ready to take part in the session
        self.session.addPlayer(self)

    def tick(self, time_passed):
        """Advances the player by the time (in seconds) specified by time_passed."""

        # Control the player
        if self.controller is not None:
            self.controller.tick(time_passed)

        # Let the tank execute any requested movement, rotation, etc.
        self.tank.tick(time_passed)

    def setController(self, controller):
        self.controller = controller
        controller.start()

    def unsetController(self):
        self.controller.stop()
        del self.controller

    def isControlled(self):
        return hasattr(self, 'controller')

    def despawn(self):
        """Removes the player from the map."""

        self.unsetController()

        # Tank parts
        self.tank.vanish()

        # Hud elements
        self.statbar.vanish()

        # Delete the references to allow GC
        del self.tank
        del self.turret
        del self.statbar

    def spawn(self):
        """Spawn the player at a location."""

        # Tank parts
        self.tank = tank.Tank(self.game, self, tank.colors[self.team % len(tank.colors)])
        self.turret = tank.Turret(self.game, self, tank.colors[self.team % len(tank.colors)])

        # Our own hud elements
        self.statbar = hudelements.StatBar(self)

        # Spawn position decided by map
        self.session.current_map.spawn(self.game, self)

    def alive(self):
        """Checks if this player is alive.
        A living player is a player that is controlled and has a health higher than zero."""

        return self.isControlled() and self.getHealth() > 0

    def damage(self, amount):
        """Do some damage to the player."""

        self.setHealth(self.getHealth() - amount)
        self.totaldamage += amount

        if (self.getHealth() <= 0):
            self.kill()

    def kill(self):
        """Instantly kill the player."""

        self.setHealth(0)
        self.deathdat = pygame.time.get_ticks()
        logger.info('Player died after %sms, took a total of %s damage', pygame.time.get_ticks() - self.createdat, self.totaldamage)

        self.despawn()

    def setHealth(self):
        pass

    def getHealth(self):
        return 0

    def isAlly(self, otherPlayer):
        """Checks if otherPlayer is an ally of this player."""

        return otherPlayer.team == self.team


class HumanPlayer(Player):
    """A human player. An InputController is automatically associated with any instance of this class."""

    def __init__(self, game, name="Human", team=0):
        Player.__init__(self, game, name, team)

        self.health = 100
        self.fuel = 100

        #from base.sprites.luminosity import Light
        #self.light = Light(game, self.turret, 30, 128, 60)
        #self.radarlayer = ComponentLayer(radar=Radar(game, self, self.turret))
        #self.maplayer = ComponentLayer(mapdiscoverer=MapDiscoverer(game, self, [self.radarlayer.radar]))

        #self.radarlayer.start()
        #self.maplayer.start()

    def tick(self, time_passed):
        Player.tick(self, time_passed)
        #for e in self.radarlayer.events.getAll():
        #    self.maplayer.event(e.getType(), e.getParams())

        #for e in self.maplayer.events.getAll():
        #    print e.getType()

    def spawn(self):
        """Spawns the player and allows interaction through input."""

        Player.spawn(self)
        self.setController(InputController(self))

    def setHealth(self, health):
        self.health = min((health, 100))

    def getHealth(self):
        return self.health

    def setFuel(self, fuel):
        self.fuel = min((fuel, 100))

    def getFuel(self):
        return self.fuel


class NonPlayer(Player):

    def __init__(self, game, name="Dummy NPC", team=0):
        Player.__init__(self, game, name, team)

    def spawn(self):
        """Spawns an NPC and controls it through components."""

        Player.spawn(self)
        self.setController(ComponentController(self, self.game))

    def setHealth(self, health):
        maxLevel = self.controller.sensors.health.maxHealth
        self.controller.sensors.health.health = min((health, maxLevel))

    def getHealth(self):
        return self.controller.sensors.health.health

    def setFuel(self, fuel):
        maxLevel = self.controller.sensors.fuel.maxFuel
        self.controller.sensors.fuel.fuel = min((fuel, maxLevel))

    def getFuel(self):
        return self.controller.sensors.fuel.fuel
