# Game Rules: Session
#
# By Gino Wuytjens, 2012

from time import time
from ..maps import map
from round import Round
from .. import events
import pygame
import math
import logging
import json

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class Session:
    """A Session consists of a number of players playing several rounds played on a sequence of maps.

    At start, the sequence of maps and number of rounds per map are determined.
    Players can be added to the session at any time."""

    game = None
    startTime = None

    maps = None
    roundsPerMap = 3
    currentRound = 0

    players = []

    def __init__(self, game, configFile):
        """Create a new session from a configuration in the specified configuration file."""

        logger.info("Creating session from file <%s>", configFile)

        self.game = game
        self.events = events.EventQueue()

        config = json.load(file(configFile))
        self.maps = config["maps"]
        self.roundsPerMap = config["roundsPerMap"]
        self.current_map = None

    def addPlayer(self, player):
        """Add a new player to the session."""

        if player not in self.players:
            self.players.append(player)
            self.events.push(events.PLAYER_JOIN, player)

    def removePlayer(self, player):
        """Remove a player from the session.
        Immediately removes the player from the current round, causing its scoreboard to become unrepresentative of the player's true performance."""

        if player in self.players:
            self.players.remove(player)
            player.kill()
            self.events.push(events.PLAYER_LEAVE, player)

    def tick(self, time_passed):
        """Advances the session with the specified time"""

        # Discard unprocessed events
        self.events.clean()

        if self.round.isFinished():
            self.currentRound += 1
            if self.currentRound % self.roundsPerMap == 0:
                self.switchMap()
            self.round = Round(self.game)
            self.round.start()
        else:
            self.current_map.tick(time_passed)
            self.round.tick(time_passed)
            for player in self.getLivePlayers():
                player.tick(time_passed)

    def switchMap(self):
        """Loads the next map in the self.maps list."""

        # Unload the current map
        if self.current_map != None:
            self.current_map.unload()
            del self.current_map

        # Load the next
        self.current_map = map.Map(self.game, self.maps[math.trunc(math.floor(self.currentRound / self.roundsPerMap) % len(self.maps))])
        self.game.screen.blit(self.current_map.surface, (0, 0))
        pygame.display.update()

    def start(self):
        self.events.push(events.SESSION_START)
        self.switchMap()
        self.startTime = time()
        self.currentRound = 1
        self.round = Round(self.game)
        self.round.start()

    def stop(self):
        self.events.push(events.SESSION_STOP)
        self.stopTime = time()
        self.round.stop()

    def findLiveItems(self, sprites):
        """Finds living items in this session that are associated with the specified list of sprites."""

        items = []
        # Check for player tanks
        for player in self.players:
            if player.tank in sprites:
                items.append(player)

        return items

    def getLivePlayers(self):

        def living(player):
            return player.alive()

        return filter(living, self.players)
