# Luminosity light group and a directional light set-able range, intensity and fov
#
# By Gino Wuytjens
import math
import pygame
from pygame.locals import *
from .. import lib


class Luminosity:
    """Luminosity is the effect of various light sources on the overall lighting of the screen."""

    def __init__(self, game, ambient=0):
        """Create a luminosity instance for the specified game and with the specified ambient light value.

        The instance will have provides the following attributes for use: game, ambient and lights.
        They represent in order the game provided to the luminosity instance, ambient light value and the list of all lights currently active."""

        self.game = game
        self.ambient = ambient
        self.lights = []

        self._drawChange = []  # This 'private' attribute is used to keep track of areas of the screen that need updating.

    def setAmbient(self, intensity):
        """Set the ambient light value. This can in turn be used by lights when drawing."""

        self.ambient = intensity

    def clear(self, surface, background):
        """Clear all lights from the screen by filling in their positions with background."""

        for light in self.lights:
            change = surface.blit(background, (light.rect.left, light.rect.top), light.rect)
            self._drawChange.append(change)

    def update(self):
        """Update the lights' statuses to prepare for drawing."""

        for light in self.lights:
            light.update()

    def draw(self, surface):
        """Draw the lights to surface."""

        for light in self.lights:
            self._drawChange.append(light.draw(surface))

        change = list(self._drawChange)
        self._drawChange = []
        return change

    def addLight(self, light):
        """Add light to the overall luminosity."""

        self.lights.append(light)

    def removeLight(self, light):
        """Remove light from the overall luminosity."""

        if light in self.lights:
            self._drawChange.append(light.rect)
            self.lights.remove(light)


class Light(pygame.sprite.Sprite):
    """Light class to illuminate the scene. Shadows are created using the walls of the current map."""

    def __init__(self, game, source, intensity, radius, fov):
        """Creates a light instance for the specified game, source is the origin of the light.
        Intensity determines the brightness, radius the overall size and fov the width of the light cone."""

        pygame.sprite.Sprite.__init__(self)  # We subclass from Sprite to use the handy spritecollide function

        self.game = game
        self.radius = radius
        self.source = source
        self.intensity = intensity
        self.width = math.radians(fov / 2)  # The 'width' is in fact half the fov and should be geometrically interpreted as how far to the left and right of the source's direction vector the light has effect
        game.luminosity.addLight(self)  # Add to the game's luminosity, so that each instance is automatically rendered after creation

        # Work with a hardware surface and palette to speed up the manipulations for shading
        self.image = pygame.Surface([self.radius * 2, self.radius * 2], HWSURFACE | SRCALPHA)
        #self.image.set_palette([[0, 0, 0], [255, 255, 255], [self.game.luminosity.ambient, self.game.luminosity.ambient, self.game.luminosity.ambient], [self.intensity, self.intensity, self.intensity]])
        self.image.set_colorkey([0, 0, 0, 0], RLEACCEL)

        # The rect must be the whole light image, because as the source's angle changes, the directional light can be anywhere inside this image
        self.rect = self.image.get_rect()
        self.rect.center = source.rect.center

    def vanish(self):
        """Remove the light from existence. It will no longer be drawn after this call."""

        self.game.luminosity.removeLight(self)

    def update(self):
        """Update the status of this light inpreparation for drawing."""

        self.rect.center = self.source.rect.center  # Stay centered on the source!

    def _getPolygon(self, x, y, box):
        """Get the light obscuring polygon from the x,y coords representing the light source position and the box that is obscuring light rays.
        This function and the tracepoint function it depends upon were based on the work by John Eriksson at http://www.pygame.org/project-Shadows-292-.html"""

        r = box.right
        l = box.left
        t = box.top
        b = box.bottom
        L = self.radius + 10

        if x >= l and x <= r:
            if y >= b:  # directly under
                tp1 = lib.gamemath.tracePoint(x, y, l, b, L)
                tp2 = lib.gamemath.tracePoint(x, y, r, b, L)
                return ((box.bottomleft, tp1, [self.radius - L, self.radius - L], [self.radius + L, self.radius - L], tp2, box.bottomright))
            else:   # directly above
                tp1 = lib.gamemath.tracePoint(x, y, l, t, L)
                tp2 = lib.gamemath.tracePoint(x, y, r, t, L)
                return ((box.topleft, tp1, [self.radius - L, self.radius + L], [self.radius + L, self.radius + L], tp2, box.topright))
        elif y >= t and y <= b:
            if x <= l:  # directly to the left
                tp1 = lib.gamemath.tracePoint(x, y, l, b, L)
                tp2 = lib.gamemath.tracePoint(x, y, l, t, L)
                return ((box.bottomleft, tp1, [self.radius + L, self.radius + L], [self.radius + L, self.radius - L], tp2, box.topleft))
            else:   # directly to the right
                tp1 = lib.gamemath.tracePoint(x, y, r, b, L)
                tp2 = lib.gamemath.tracePoint(x, y, r, t, L)
                return ((box.bottomright, tp1, [self.radius - L, self.radius + L], [self.radius - L, self.radius - L], tp2, box.topright))
        elif y <= t:
            if x <= l:  # upper left
                tp1 = lib.gamemath.tracePoint(x, y, r, t, L)
                tp2 = lib.gamemath.tracePoint(x, y, l, b, L)
                return ((box.topleft, box.topright, tp1, tp2, box.bottomleft))
            else:     # upper right
                tp1 = lib.gamemath.tracePoint(x, y, l, t, L)
                tp2 = lib.gamemath.tracePoint(x, y, r, b, L)
                return ((box.topright, box.topleft, tp1, tp2, box.bottomright))
        else:
            if x <= l:  # lower left
                tp1 = lib.gamemath.tracePoint(x, y, r, b, L)
                tp2 = lib.gamemath.tracePoint(x, y, l, t, L)
                return ((box.bottomleft, box.bottomright, tp1, tp2, box.topleft))
            else:     # lower right
                tp1 = lib.gamemath.tracePoint(x, y, l, b, L)
                tp2 = lib.gamemath.tracePoint(x, y, r, t, L)
                return ((box.bottomright, box.bottomleft, tp1, tp2, box.topright))

        return None

    def get_bounding_rect(self):
        """An analogue to the pygame.Surface.get_bounding_rect() function, returns the smallest rectangle fully encompassing the light cone."""

        rect = self.image.get_bounding_rect()
        rect.inflate_ip(5, 5)  # The light bounding rect is increased with one pixel in each direction. Shadows would otherwise eliminate the light cast on the edge of walls.
        rect.left += self.rect.left
        rect.top += self.rect.top
        return rect

    def draw(self, surface):
        """Draw this light to surface."""

        self.image.fill([0, 0, 0, 0])  # Clear our internal representation

        # Draw the unobscured light
        pygame.draw.arc(self.image, [self.intensity, self.intensity, self.intensity, 255], self.image.get_rect(), self.source.getAngle() - self.width, self.source.getAngle() + self.width, self.radius)

        # Find obscuring rectangles
        nrects = []
        for r in self.game.session.current_map.getWalls():
            if self.rect.colliderect(r):  # If the wall collides with our rectangle
                nr = r.clip(self.rect)  # We only need the area inside our rectangle
                nr.top = nr.top - self.rect.top
                nr.left = nr.left - self.rect.left
                nrects.append(nr)

        for r in nrects:  # Get the polygonal shadow for each of the obscuring rectangles
            p = self._getPolygon(self.radius, self.radius, r)
            if p:
                pygame.draw.polygon(self.image, [0, 0, 0, 0], p, 0)

        return surface.blit(self.image, self.rect.topleft, special_flags=BLEND_ADD)  # We use BLEND_ADD to increase the brightness of the surface where the light is present
