# Pickup, class that represents a sprite that can be 'picked up' by players and has some effect afterwards
#
# By Gino Wuytjens, 2012

import pygame


class Pickup(pygame.sprite.Sprite):
    """Represents a Pickup or powerup that is picked up and subsequently has an effect on the player."""

    def __init__(self, spritesheet, game, imageID, position):
        """Creates a new pickup. Pickups of this class will not have an effect on players.
        imageID represents the name of the image to load from the active spritesheet."""
        pygame.sprite.Sprite.__init__(self, game.pickups)

        self.image = spritesheet.getSpriteImage(imageID)
        self.rect = self.image.get_rect()
        self.rect.center = position
        self.game = game
        self.reappearingTime = 0

    def getPos(self):
        """Returns the pickup's position."""

        return self.rect.center

    def tick(self, time_passed):
        """Advance the time by time_passed.
        This checks if any player is currently close enough to pick us up.
        If there is such a player, self.affectPlayer(player) is called."""

        # Affect players if we are active only
        if self.isActive():
            for player in self.game.session.getLivePlayers():
                if self.rect.colliderect(player.tank.rect):
                    self.affectPlayer(player)
                    break
        elif self.reappearingTime > 0:
            self.reappearingTime -= time_passed
        else:
            self.setActive()

    def setActive(self, active=True):
        """Sets the pickup as active (or not, if the optional parameter active is set to False)."""

        if active:
            self.game.pickups.add(self)
        else:
            self.vanish()

    def isActive(self):
        """Returns True if the current pickup is active (can be picked up)."""
        return len(self.groups()) > 0

    def vanish(self, reappear=0):
        """Removes the sprite from any groups it belongs to, thus causing it to vanish from the screen.
        The optional reappear parameter specifies the number of milliseconds it will take for this pickup to reappear."""

        self.reappearingTime = reappear
        for group in self.groups():
            group.remove(self)

    def update(self, time_passed):
        """This method should be overridden by subclasses if they which to alter the graphical appearance of the pickup."""

        pass

    def affectPlayer(self, player):
        """This method should be overridden by subclasses to affect players that pick up the pickup."""

        pass


class FuelBarrel(Pickup):
    """A fuel barrel will replenish fuel."""

    def __init__(self, spritesheet, game, position):
        Pickup.__init__(self, spritesheet, game, "barrel", position)

    def affectPlayer(self, player):
        """Tops up the player's fuel."""

        player.setFuel(999)
        self.vanish(5000)  # 5 seconds to respawn


class RepairShop(Pickup):
    """A repair shop will restore health."""

    def __init__(self, spritesheet, game, position):
        Pickup.__init__(self, spritesheet, game, "garage", position)

    def affectPlayer(self, player):
        """Tops up the player's health."""

        player.setHealth(999)
        self.vanish(5000)  # 5 seconds to respawn
