# Sprite module, contains base classes for sprites
#
# Gino Wuytjens

import math
import logging
import pygame
from ..lib import gamemath

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class OwnedObject:

    def __init__(self, owner):
        self.owner = owner

    def getOwner(self):
        return self.owner

    def setOwner(self, owner):
        self.owner = owner

    def isMine(self, other):
        return isinstance(other, OwnedObject) and self.owner is other.owner


class MoveableSprite(pygame.sprite.Sprite):

    NORMAL_SPEED = 20

    # Creates a movable sprite from a bounding box and world to move within.
    # The bounding box represent the actual sprites location and size
    # The image (used by subclasses of this class) representing the sprite may have a position derived from this bounding box
    # and the image size may be different than the bounding box's width and length attributes specify
    def __init__(self, bounding_box, world, containers):
        pygame.sprite.Sprite.__init__(self, containers)
        self.bbox = bounding_box  # bounding box is a pygame.Rect ((left edge position, top edge position), (width, height))
        self.x = bounding_box.centerx
        self.y = bounding_box.centery
        self.speed = 0.0  # the speed the sprite is travelling at, in pixels/second
        self.turnSpeed = 0.0  # the speed with which the sprite is turning
        self.angle = 0.0
        self.ax = 1.0  # angle we're headed in projected on the x-axis, see setAngle(alpha)
        self.ay = 0.0  # angle we're headed in projected on the y-axis, see setAngle(alpha)
        self.world = world
        (self.worldwidth, self.worldheight) = self.world.area  # x and y position limits (limit is 1 greater than allowed maximum coordinate)
        self.movingToPos = (-1.0, -1.0)
        self.turnToAngle = None

    # Sets this sprite's position to the given coordinates
    def setPos(self, (x, y)):
        distance = gamemath.distance((self.x, self.y), (x, y))
        self.bbox.center = (x, y)
        self.x = x
        self.y = y

        #print self.movingToPos
        if self.movingToPos != (-1.0, -1.0) and self.speed != 0.0:
            if self.isNear(self.movingToPos):
                print '[D] Destination reached, stopping.'
                self.movingToPos = (-1.0, -1.0)
                self.setSpeed(0.0)

        return distance

    # Sets the angle this sprite is oriented in
    # alpha - angle in radians, 0 is vector <0,1>
    def setAngle(self, alpha):
        sane_angle = alpha % (2.0 * math.pi)  # prevents the angle value from rising to silly levels
        if (self.angle != sane_angle):
            self.ax = math.cos(sane_angle)
            self.ay = math.sin(sane_angle)
            self.angle = sane_angle

    # Gets the angle this sprite is oriented in
    # Returns an angle in radians
    def getAngle(self):
        return self.angle

    # Gets the angle this sprite is oriented in
    # Returns the projections of the angle onto x and y axis
    def getAngleProjection(self):
        return (self.ax, self.ay)

    # Sets the spead the sprite is travelling at
    def setSpeed(self, speed):
        self.speed = speed

    def setTurnSpeed(self, turnSpeed):
        self.turnSpeed = turnSpeed

    def getTurnSpeed(self):
        return self.turnSpeed

    # Gets the position of this sprite, center coordinates are returned
    def getPos(self):
        return self.bbox.centerx, self.bbox.centery

    # Gets the bounding box of this sprite
    def getBoundingBox(self):
        return self.bbox

    # Gets the x and y directional speeds
    def getSpeed(self):
        return self.speed

    # Updates the coordinates
    def updatePosition(self, time_passed):
        # Adjust angle as specified by turnspeed
        nextAngle = self.getAngle() + self.turnSpeed
        if self.turnToAngle != None and ((self.turnSpeed > 0 and nextAngle > self.turnToAngle) or (self.turnSpeed < 0 and nextAngle < self.turnToAngle)):
            self.setAngle(self.turnToAngle)
            self.turnToAngle = None
            self.turnSpeed = 0.0
        else:
            self.setAngle(self.getAngle() + self.turnSpeed)

        # determine our nex coordinates, but do not assign them!
        x = self.x + self.speed * self.ax * time_passed / 1000.0
        y = self.y + self.speed * -self.ay * time_passed / 1000.0
        # let the setPos((x,y)) function set the coordinates, subclasses may then override this method and add new behaviour like collision
        # To support subclasses having return values (i.e. for reason of failure), return self.setPos return value here, even though we don't have a return value for that function
        return self.setPos((x, y))

    def turnTo(self, angle, turnSpeed):
        self.setTurnSpeed(turnSpeed)
        self.turnToAngle = angle

    def moveTo(self, sprite):
        self.turnTo(sprite)
        self.setSpeed(self.NORMAL_SPEED)
        self.movingToPos = (sprite.x, sprite.y)

    def isNear(self, (x, y)):
        return gamemath.distance((self.x, self.y), (x, y)) < 10

    def isMovingToDestination(self):
        return self.movingToPos != (-1.0, -1.0)

    def getClosest(self, sprites):
        """Returns the closest, in euclidean distance, sprite from a list of sprites."""

        nrSprites = len(sprites)
        if nrSprites == 0:
            return None
        elif nrSprites == 1:
            return sprites[0]
        else:
            shortestDistance = 99999
            closestSprite = None
            for sprite in sprites:
                distance = gamemath.distanceSprite(self, sprite)
                if distance < shortestDistance:
                    shortestDistance = distance
                    closestSprite = sprite
            return closestSprite


class CollideableSprite(MoveableSprite):

    # Creates a Collideable Sprite from the collisionmesh it must collide with and a bounding box (pygame.Rect) for the object itself
    def __init__(self, bounding_box, world, containers):
        MoveableSprite.__init__(self, bounding_box, world, containers)

    # Sets the x and y coordinates to the supplied x and y values.
    # If x and y values are outside of world or colliding, they are clipped to the closest edge
    # returns a tuple of (succes - boolean, reason of failure - string)
    def setPos(self, (x, y)):
        # Set the coordinates to allow pygame's sprite methods to be used in collision tests
        backupPosition = self.getPos()
        distance = MoveableSprite.setPos(self, (x, y))

        if not self.__clip(x, y):
            # Not clipping outside the map
            (collision, colliders) = self.world.collision(self)  # self.__collide()
            if collision != "none":
                # Colliding with something
                MoveableSprite.setPos(self, backupPosition)
                logger.debug("%s,%s collides with %s", x, y, colliders)
                return (collision, colliders, 0)
            else:
                logger.debug("%s,%s is ok", x, y)
                return ("ok", [], distance)
        else:  # Clipping, reset to previous position
            MoveableSprite.setPos(self, backupPosition)
            logger.debug("%s,%s clips", x, y)
            return ("clip", [], 0)

    # Clips a coordinate to the given limit
    def __clip(self, x, y):
        left = x - self.bbox.width / 2
        right = x + self.bbox.width / 2

        top = y - self.bbox.height / 2
        bottom = y + self.bbox.height / 2

        clip = False

        if (left < 0) or (right >= self.worldwidth):
            if (gamemath.distance_points(left, 0) > gamemath.distance_points(right, self.worldwidth)):
                x = self.bbox.width / 2
            else:
                x = self.worldwidth - self.bbox.width / 2
            clip = True
        if (top < 0) or (bottom >= self.worldheight):
            if (gamemath.distance_points(top, 0) > gamemath.distance_points(bottom, self.worldheight)):
                y = self.bbox.height / 2
            else:
                y = self.worldheight - self.bbox.height / 2
            clip = True
        return clip

    # Checks for collisions if no collisions are found, the provided bbox is set
    def __collide(self):
        return self.world.collision(self)
