# Tank
#
# Gino Wuytjens

import logging
import math
import pygame
from pygame.locals import *

import sprite
import projectile
from .. import lib

logger = logging.getLogger(__name__)

colors = ['blue', 'green', 'pink', 'cyan', 'yellow', 'red']


class Tombstone(pygame.sprite.Sprite):

    def __init__(self, position, map, containers):
        pygame.sprite.Sprite.__init__(self, containers)
        self.image = map.getSpritesheet().getSpriteImage('flag white')
        self.rect = self.image.get_rect()
        self.rect.center = position


class Turret(sprite.OwnedObject):
    """Turret for a tank. Can turn 360 degrees and shoot."""

    def __init__(self, game, owner, colour):
        """Creates a turret of a certain colour, belonging to the specified owner.
        The owner will handle rendering the self.image attribute to the screen."""

        sprite.OwnedObject.__init__(self, owner)

        self.turret_img = self.owner.game.session.current_map.spritesheet.getSpriteImage('turret ' + colour)  # Load the turret img with the requested colour
        self.image = self.turret_img
        self.angle = 0.0
        self.turnSpeed = 0.0
        self.turnToAngle = None
        self.owner.tank.turret = self
        self.rect = self.owner.tank.rect
        self.imagechange = True  # Force rendering the sprite

    def getPos(self):
        """Returns the position of the turret as a tuple (x, y)."""

        return self.rect.center

    def setAngle(self, angle):
        """Set the turret's angle to the specified value.
        The angle value is absolute (not relative to the owner's angle)."""

        sane_angle = angle % (2.0 * math.pi)  # prevents the angle value from rising to silly levels
        if (self.angle != sane_angle):
            self.angle = sane_angle
            self.imagechange = True

    def setTurnSpeed(self, turnSpeed):
        self.turnSpeed = turnSpeed

    def getTurnSpeed(self):
        return self.turnSpeed

    def getAngle(self):
        """Returns the angle the turret is currently looking at."""

        return self.angle

    def update(self, time_passed):
        """Update the turret and advance the time by time_passed seconds."""

        nextAngle = self.getAngle() + self.turnSpeed
        if self.turnToAngle != None and ((self.turnSpeed > 0 and nextAngle > self.turnToAngle) or (self.turnSpeed < 0 and nextAngle < self.turnToAngle)):
            self.setAngle(self.turnToAngle)
            self.turnToAngle = None
            self.turnSpeed = 0.0
        else:
            self.setAngle(self.getAngle() + self.turnSpeed)

        if self.imagechange:
            self.refreshImage()

    def turnTo(self, angle, turnSpeed):
        self.setTurnSpeed(turnSpeed)
        self.turnToAngle = angle

    def refreshImage(self):
        #self.image = pygame.transform.rotate(self.turret_img, self.angle*180/math.pi) # rotate from fresh sprite to avoid artifacts
        self.image = lib.gamemath.rotate(self.turret_img, self.angle * 180 / math.pi)
        self.owner.tank.refreshImage()
        self.imagechange = False

    # fire a bullet
    def fire(self):
        projectile.Bullet(self.owner)


class Tank(sprite.OwnedObject, sprite.CollideableSprite):

    def __init__(self, game, player, colour):
        sprite.OwnedObject.__init__(self, player)
        self.game = game
        self.body_img = game.session.current_map.spritesheet.getSpriteImage('tank ' + colour)  # Load the tank img with the requested colour
        self.image = self.body_img  # Before we enter the world, the default orientation is used
        sprite.CollideableSprite.__init__(self, pygame.Rect((-100, -100), self.body_img.get_size()), game.session.current_map, game.playerSprites)
        self.rect = self.bbox  # set positioning rect to our bbox
        self.imagechange = True  # force refreshing our image on first frame

        #self.turret = Turret('blue', self)
        #self.light = luminosity.Light(game, self, 30)
        #self.actor = actorobj if actorobj != None else actor.Actor(self)

        #self.boundingbox = hudelements.BoundingBox(self)

    def tick(self, time_passed):
        """Advance the time by time_passed seconds."""

        if self.owner.getFuel() > 0:
            (status, colliders, distance) = self.updatePosition(time_passed)
            self.owner.setFuel(self.owner.getFuel() - distance * 0.1)

        self.turret.update(time_passed)

    # Update position etc.
    def update(self, time_passed):

        if self.imagechange:
            self.refreshImage()

    # Sets the angle the tank body is oriented at, in radians
    def setAngle(self, alpha):
        previous = self.angle
        sprite.CollideableSprite.setAngle(self, alpha)
        difference = alpha - previous
        self.turret.setAngle(self.turret.getAngle() + difference)  # make the turret rotate with the tank
        self.imagechange = True

    # Sets the turret angle
    def setTurretAngle(self, alpha):
        self.turret.setAngle(alpha)

    # Gets the turret angle
    def getTurretAngle(self):
        return self.turret.getAngle()

    def turnTurretTo(self, object):
        assert isinstance(object, sprite.MoveableSprite), '[E] Tank cannot turn turret at anything but other MoveableSprites'
        angle = lib.gamemath.angle((self.x, self.y), (object.x, object.y))
        self.setTurretAngle(angle)

    # Gets our turret
    def getTurret(self):
        return self.turret

    # Refreshes the image attribute to represent the whole tank
    def refreshImage(self):
        #enter = self.rect.center
        # The tank body first
        #self.image = pygame.transform.rotate(self.body_img, self.angle*180/math.pi)
        self.image = lib.gamemath.rotate(self.body_img, self.angle * 180 / math.pi)
        self.mask = pygame.mask.from_surface(self.image)
        # And then the turret on top
        self.image.blit(self.turret.image, (0, 0))

        #self.rect = self.image.get_rect()
        #self.rect.center = enter
        self.imagechange = False

    def setActor(self, actor):
        del self.actor
        self.actor = actor

    def vanish(self):
        #Tombstone(self.rect.center, self.current_map, self.groups())
        #self.turret.vanish()
        #self.light.vanish()
        #self.boundingbox.vanish()
        for group in self.groups():
            group.remove(self)

    def fire(self):
        self.turret.fire()

    def isEnemy(self, otherTank):
        """Checks if otherTank is hostile towards this tank."""

        return not self.owner.isAlly(otherTank.owner)

    def isFriendly(self, otherTank):
        """Checks if otherTank is friendly towards this tank (and not this tank itself!)."""

        return not otherTank is self and self.owner.isAlly(otherTank.owner)
