# Game Testbed
#
# Gino Wuytjens

# Standard Libraries
import logging
import json
import argparse
import cProfile

# 3rd Party libraries
import pygame
from pygame.locals import *

# Game components
from base.players.player import HumanPlayer
from base.players.player import NonPlayer
from base.sprites.luminosity import Luminosity
from base.rules.session import Session

logging.basicConfig()
logger = logging.getLogger()
logger.setLevel(logging.DEBUG)


class Game:
    """The main entrypoint for the game."""

    # Initialization tasks
    def __init__(self):
        """Create a new game"""
        logger.info("Creating game")

        # Read configuration file
        self.settings = json.load(file("game.cfg"))

        # Initialize Pygame library
        pygame.init()
        pygame.display.set_caption('Statechart AI Testbed')
        self.clock = pygame.time.Clock()
        self.screen = pygame.display.set_mode((self.settings["resolution"][0], self.settings["resolution"][1]), HWSURFACE | DOUBLEBUF)
        self.schedule = {}

        # Sprite groups
        self.playerSprites = pygame.sprite.OrderedUpdates()  # Sprites belonging to players
        self.pickups = pygame.sprite.RenderUpdates()
        #self.projectileSprites = pygame.sprite.RenderUpdates() # Sprites belonging to projectiles
        self.hud = pygame.sprite.RenderUpdates()  # Sprites belonging to the hud
        self.luminosity = Luminosity(self)  # Lighting (helps by showing view range of npcs)

    def after(self, time, function):
        self.schedule[function] = time

    def runAfters(self, time_passed):
        stoppedTimers = []
        for function, time_left in self.schedule.iteritems():
            self.schedule[function] = time_left - time_passed
            if self.schedule[function] <= 0:
                stoppedTimers.append(function)

        for function in stoppedTimers:
            del self.schedule[function]
            function()

    def start(self, sessionConfigFile):
        # Create a session
        self.session = Session(self, sessionConfigFile)

        # Make some players
        player1 = NonPlayer(self, team=1)
        player2 = NonPlayer(self, team=2)
        # player3 = NonPlayer(self, team=3)
        # player4 = NonPlayer(self, team=4)
        # player5 = NonPlayer(self, team=5)
        # player6 = NonPlayer(self, team=6)
        # player7 = NonPlayer(self, team=7)
        # player8 = NonPlayer(self, team=8)

        # Start the session
        self.session.start()

        # Show the screen
        pygame.display.update()
        # Main loop
        while 1:

            # Frame Rate Limit
            time_passed = self.clock.tick(40)

            # Handle events since previous frame
            if len(pygame.event.get(QUIT)) > 0:
                logger.info("FPS was %s", self.clock.get_fps())
                return

            # Run any scheduled functions
            self.runAfters(time_passed)

            # Advance the time for all behavior and game rule logic
            self.session.tick(time_passed)

            # Anything below is rendering and pygame bookkeeping

            # Redraw all visual elements
            self.screen.blit(self.session.current_map.surface, (0, 0))

            self.luminosity.clear(self.screen, self.session.current_map.surface)
            self.playerSprites.clear(self.screen, self.session.current_map.surface)
            self.pickups.clear(self.screen, self.session.current_map.surface)
            self.hud.clear(self.screen, self.session.current_map.surface)

            self.playerSprites.update(time_passed)
            self.pickups.update(time_passed)
            self.luminosity.update()
            self.hud.update(time_passed)

            # clean up events
            pygame.event.get()

            dirtyPickupRegions = self.pickups.draw(self.screen)
            dirtyPlayerRegions = self.playerSprites.draw(self.screen)
            dirtyLightRegions = self.luminosity.draw(self.screen)
            dirtyHudRegions = self.hud.draw(self.screen)

            pygame.display.update(dirtyPickupRegions)
            pygame.display.update(dirtyPlayerRegions)
            pygame.display.update(dirtyLightRegions)
            pygame.display.update(dirtyHudRegions)
            #pygame.display.update()


def main():
    g = Game()
    parser = argparse.ArgumentParser()
    parser.add_argument('-s', '--session', default="defaultsession.cfg")
    args = parser.parse_args()
    g.start(args.session)

# main
if __name__ == "__main__":
    main()
